/***************************************************************************
 * jtalk -- talk-like Jabber client
 * (C) Nicolas George - 2006
 * jtalkd.c: main loop of the connection daemon
 ***************************************************************************/

/*
  This file is part of jtalk.

  jtalk is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  jtalk is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with jtalk; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <glib.h>
#include <loudmouth/loudmouth.h>

#include "jtalk.h"

typedef struct jtalk_peer_s jtalk_peer_t;
struct jtalk_peer_s {
    jtalk_connection_t *connection;
    jtalk_contact_t *contact;
    jtalk_client_t *client;
    void (*handler)(jtalk_peer_t *, jtalk_message_t *);
};

GList *connections = NULL;
GList *clients = NULL;
GList *peers = NULL;

void
jtalk_print_timestamp(jtalk_timestamp_t t, char *ts)
{
    time_t now;
    struct tm *tm;

    now = t / 1000;
    tm = localtime(&now);
    sprintf(ts, "%02d:%02d:%02d", tm->tm_hour, tm->tm_min, tm->tm_sec);
}

static int
split_words(char *t, char **w)
{
    int n = 0;

    if(strlen(t) >= 6 && memcmp(t, "/argv ", 6) == 0) {
	/* this is somewhat a hack to put that here, but not much; TODO: add
	 * a full quoting capability */
	if(w != NULL) {
	    t[5] = 0;
	    w[0] = t;
	    w[1] = t + 6;
	}
	return(2);
    }
    while(1) {
	while(*t == ' ' || *t == '\r')
	    t++;
	if(*t == 0)
	    break;
	n++;
	if(w != NULL)
	    *(w++) = t;
	while(*t != 0 && *t != ' ' && *t != '\t')
	    t++;
	if(w != NULL && *t != 0)
	    *(t++) = 0;
    }
    return(n);
}

static char *
sanitize_text(const unsigned char *text)
{
    const unsigned char *e, *p;
    static GString *s = NULL;

    if(s == NULL)
	s = g_string_new(NULL);
    else
	g_string_truncate(s, 0);
    while(*text == ' ' || *text == '\t' || *text == '\r' || *text == '\n')
	text++;
    e = text + strlen(text);
    while(e > text && (e[-1] == ' ' || e[-1] == '\t' || e[-1] == '\r' ||
	e[-1] == '\n'))
	e--;
    while(text < e) {
	for(p = text; p < e && (*p == '\n' || (*p >= ' ' && *p != 0x7F)); p++);
	if(p > text) {
	    g_string_append_len(s, text, p - text);
	    text = p;
	} else {
	    if(*text == '\t')
		g_string_append_c(s, ' ');
	    else if(*text != '\t')
		g_string_append_printf(s, "^%c", *text ^ 64);
	    text++;
	}
    }
    return(g_strdup(s->str));
}

jtalk_timestamp_t
jtalk_now(void)
{
    struct timeval tv;

    gettimeofday(&tv, NULL);
    return((jtalk_timestamp_t)tv.tv_sec * 1000 + tv.tv_usec / 1000);
}

void
jtalk_client_printf(jtalk_client_t *client, const char *fmt, ...)
{
    va_list args;
    gchar *t;

    va_start(args, fmt);
    t = g_strdup_vprintf(fmt, args);
    va_end(args);
    if(client == NULL) {
	jtalk_tell_user_direct_and_free(0, t);
    } else {
	socket_write(client, t);
	g_free(t);
    }
}

static void
jtalk_client_print_message(jtalk_client_t *client, jtalk_message_t *m)
{
    char tst[9];
    const char *hname, *from;

    jtalk_print_timestamp(m->timestamp, tst);
    if(m->from == NULL) {
	from = "self"; /* XXX */
	hname = jtalk_esc_self_name;
    } else {
	from = m->from->jid;
	hname = jtalk_esc_name;
    }
    jtalk_client_printf(client, "%s %s<%s>%s %s",
	tst, hname, from, jtalk_esc_reset, m->text);
}

static void
jtalk_client_on_message(jtalk_peer_t *p, jtalk_message_t *m)
{
    jtalk_client_print_message(p->client, m);
}

static jtalk_connection_t *
jtalk_connection_choose_for_contact(jtalk_client_t *cli, jtalk_contact_t *ct)
{
    jtalk_connection_t *best = NULL, *c;
    int best_score = 0, score;
    GList *cp, *mp;
    jtalk_message_t *m;
    int have_msg, have_und;
    int n_cand = 0;
    const char *bjid;

    if(cli->preferred_connection != NULL)
	return(cli->preferred_connection);
    for(cp = connections; cp != NULL; cp = cp->next) {
	c = cp->data;
	score = 0;
	if(c->connected)
	    score += jtalk_opt_connection_choose_connected;
	have_msg = 0;
	have_und = 0;
	for(mp = ct->msg_last; mp != NULL; mp = mp->prev) {
	    m = mp->data;
	    if(m->connection == c) {
		have_msg = 1;
		if(!m->delivered) {
		    have_und = 1;
		    break;
		}
	    }
	}
	if(have_msg)
	    score += jtalk_opt_connection_choose_messages;
	if(have_und)
	    score += jtalk_opt_connection_choose_undelivered;
	if(score > best_score) {
	    best = c;
	    best_score = score;
	}
	n_cand++;
    }
    if(n_cand > 1) {
	/* TODO */
	bjid = best->jid;
	if(bjid == NULL)
	    bjid = g_strdup_printf("%s@%s", best->user, best->server);
	jtalk_client_printf(cli, "Talking to %s<%s>%s as %s<%s>%s.",
	    jtalk_esc_name, ct->jid, jtalk_esc_reset,
	    jtalk_esc_self_name, bjid, jtalk_esc_reset);
	if(best->jid == NULL)
	    free((char *)bjid);
    }
    return(best);
}

static void
jtalk_add_peer(jtalk_client_t *client, char *name)
{
    jtalk_contact_t *ct;
    jtalk_peer_t *p;

    ct = jtalk_contact_find(name, 1);
    p = g_new(jtalk_peer_t, 1);
    p->connection = jtalk_connection_choose_for_contact(client, ct);
    p->contact = ct;
    p->client = client;
    p->handler = jtalk_client_on_message;
    peers = g_list_append(peers, p);
}

static void
jtalk_peer_free(jtalk_peer_t *peer)
{
    g_free(peer);
}

static gint
jtalk_message_compare_ts(gconstpointer a_v, gconstpointer b_v)
{
    const jtalk_message_t *a = a_v, *b = b_v;

    return(a->timestamp < b->timestamp ? -1 :
	a->timestamp > b->timestamp ? 1 : 0);
}

static void
jtalk_send_old_mesages(jtalk_client_t *client)
{
    GList *peer_p, *mp;
    jtalk_peer_t *peer;
    GSList *old_messages, *msg_p;
    jtalk_message_t *msg, *prev_msg;
    jtalk_timestamp_t older = 0;

    old_messages = NULL;
    for(peer_p = peers; peer_p != NULL; peer_p = peer_p->next) {
	peer = peer_p->data;
	if(peer->client != client)
	    continue;
	for(mp = peer->contact->msg_first; mp != NULL; mp = mp->next)
	    old_messages = g_slist_prepend(old_messages, mp->data);
    }
    old_messages = g_slist_sort(old_messages, jtalk_message_compare_ts);
    prev_msg = NULL;
    for(msg_p = old_messages; msg_p != NULL; msg_p = msg_p->next) {
	msg = msg_p->data;
	older = msg->timestamp;
	if(!msg->delivered)
	    break;
    }
    older -= jtalk_opt_deliver_old_messages;
    for(msg_p = old_messages; msg_p != NULL; msg_p = msg_p->next) {
	msg = msg_p->data;
	if(msg->timestamp < older)
	    continue;
	if(msg != prev_msg)
	    jtalk_client_print_message(client, msg);
	if(!msg->delivered) {
	    g_assert(msg->from != NULL);
	    msg->from->undelivered_msg--;
	    msg->delivered = 1;
	}
	prev_msg = msg;
    }
    g_slist_free(old_messages);
}

static void
jtalk_add_account(char **a, int na)
{
    jtalk_connection_t *c;

    if((c = jtalk_connection_new(a, na)) == NULL)
	return;
    connections = g_list_append(connections, c);
}

static void
jtalk_list_contacts(jtalk_client_t *client, char **a, int na)
{
    GList *contacts, *cp;
    jtalk_contact_t *c;
    int tech = 0;

    while(na > 0) {
	if(strcmp(*a, "tech") == 0) {
	    tech = 1;
	    na--;
	    a++;
	} else {
	    jtalk_client_printf(client, "%sError: /who: %s: unknown option.%s",
		jtalk_esc_error, *a, jtalk_esc_reset);
	    return;
	}
    }
    contacts = jtalk_contact_get_all();
    contacts = jtalk_contact_sort(contacts);
    for(cp = contacts; cp != NULL; cp = cp->next) {
	c = cp->data;
	if(tech) {
	    jtalk_client_printf(client, "%s%s:%s",
		c->available_res == NULL ? "" : "O",
		c->undelivered_msg == 0 ? "" : "U",
		c->jid);
	} else {
	    if(c->available_res == NULL && c->undelivered_msg == 0)
		continue;
	    jtalk_client_printf(client, "%s %s %s",
		c->available_res == NULL ? jtalk_esc_offline : jtalk_esc_online,
		c->undelivered_msg == 0 ?
		    jtalk_esc_delivered : jtalk_esc_undelivered,
		c->jid);
	}
    }
    g_list_free(contacts);
}

static void
jtalk_set_from(jtalk_client_t *client, char **a, int na)
{
    jtalk_connection_t *c;
    GList *cp;
    char *cjid;
    int found;

    client->preferred_connection = NULL;
    if(na == 0)
	return;
    for(cp = connections; cp != NULL; cp = cp->next) {
	c = cp->data;
	if(c->jid == NULL) {
	    /* TODO: I have to eliminate this reconstruction of jid */
	    cjid = g_strdup_printf("%s@%s", c->user, c->server);
	    found = strcmp(a[0], cjid);
	    g_free(cjid);
	} else {
	    found = strcmp(a[0], c->jid);
	}
	if(found == 0) {
	    client->preferred_connection = c;
	    return;
	}
    }
    jtalk_client_printf(client, "%sError: /from: %s not found.%s",
	jtalk_esc_error, a[0], jtalk_esc_reset);
}

void
jtalk_command_from_client(jtalk_client_t *client, unsigned char *line)
{
    int n = split_words(line, NULL);
    char *word[n];
    split_words(line, word);

    if((n == 1 || n == 2) && strcmp(word[0], "/argv") == 0) {
	if(n == 2) {
	    if(word[1][0] == '/')
		jtalk_command_from_client(client, word[1]);
	    else
		jtalk_add_peer(client, word[1]);
	} else {
	    jtalk_send_old_mesages(client);
	}
    } else if(n > 1 && strcmp(word[0], "/account") == 0) {
	jtalk_add_account(word + 1, n - 1);
    } else if(n == 1 && strcmp(word[0], "/quit") == 0) {
	g_source_remove(client->source_id);
    } else if(n >= 1 && strcmp(word[0], "/who") == 0) {
	jtalk_list_contacts(client, word + 1, n - 1);
    } else if(n >= 1 && strcmp(word[0], "/sorter") == 0) {
	jtalk_sorter_command(client, word + 1, n - 1);
    } else if(n >= 1 && strcmp(word[0], "/from") == 0) {
	jtalk_set_from(client, word + 1, n - 1);
    } else {
	jtalk_client_printf(client, "%sError: %s: unknown command.%s",
	    jtalk_esc_error, word[0], jtalk_esc_reset);
    }
}

static void
jtalk_send_to_peers(jtalk_client_t *client, char *line)
{
    GList *pp;
    jtalk_peer_t *p;
    jtalk_message_t *m;

    m = jtalk_message_new(NULL, line);
    m->delivered = 1;
    for(pp = peers; pp != NULL; pp = pp->next) {
	p = pp->data;
	if(p->client != client)
	    continue;
	jtalk_connection_send_chat_message(p->connection, p->contact, line);
	jtalk_contact_add_message(p->contact, m);
    }
    jtalk_log_message(m);
    jtalk_client_print_message(client, m);
}

void
jtalk_line_from_client(jtalk_client_t *client, unsigned char *line)
{
    if(!g_utf8_validate(line, -1, NULL)) {
	jtalk_client_printf(client, "%sError: not UTF-8, fix your client",
	    jtalk_esc_error, jtalk_esc_reset);
	return;
    }
    if(line[0] == '/') {
	if(line[1] == '/') {
	    line++;
	} else {
	    jtalk_command_from_client(client, line);
	    return;
	}
    }
    jtalk_send_to_peers(client, line);
}

static void
jtalk_notify_chat_text(jtalk_message_t *m)
{
    char tst[9];
    char *e;
    int l = 72;
    int value;

    jtalk_print_timestamp(m->timestamp, tst);
    e = strchr(m->text, '\n');
    if(e != NULL && e < m->text + l)
	l = e - m->text;
    value = jtalk_sorter_eval_l(5, "type", "unconnected message",
	"from", m->from->jid, "text", m->text, NULL);
    jtalk_tell_user(value, "(%s) message from %s<%s>%s:\n%.*s",
	tst, jtalk_esc_name, m->from->jid, jtalk_esc_reset, l, m->text);
}

void
jtalk_got_chat_text(jtalk_connection_t *con, const char *from,
    const char *text)
{
    jtalk_contact_t *ct;
    int handled = 0;
    GList *pp;
    jtalk_peer_t *p;
    jtalk_message_t *m;

    text = sanitize_text(text);
    m = jtalk_message_new(from, text);
    m->connection = con;
    jtalk_log_message(m);
    ct = m->from;
    for(pp = peers; pp != NULL; pp = pp->next) {
	p = pp->data;
	if(p->connection == con && p->contact == ct) {
	    p->handler(p, m);
	    handled++;
	}
    }
    if(handled == 0) {
	jtalk_notify_chat_text(m);
	m->from->undelivered_msg++;
    } else {
	m->delivered = 1;
    }
    g_free((char *)text);
}

void
jtalk_client_init(jtalk_client_t *client)
{
    client->preferred_connection = NULL;
    clients = g_list_append(clients, client);
}

void
jtalk_client_closed(jtalk_client_t *client)
{
    GList *pp, *npp;
    jtalk_peer_t *p;

    for(pp = peers; pp != NULL; pp = npp) {
	npp = pp->next;
	p = pp->data;
	if(p->client == client) {
	    jtalk_peer_free(p);
	    peers = g_list_delete_link(peers, pp);
	    /* TODO: delete_link is linear */
	}
    }
    g_free(client->buffer);
    g_free(client);
}

void
jtalk_got_roster_contact(const char *jid, int subs_from, int subs_to)
{
    jtalk_contact_t *c;

    c = jtalk_contact_find(jid, 1);
    c->subscribed_from = c->subscribed_from || subs_from;
    c->subscribed_to = c->subscribed_to || subs_to;
}
