/***************************************************************************
 * jtalk -- talk-like Jabber client
 * (C) Nicolas George - 2006
 * socket.c: socket interface to the front-ends
 ***************************************************************************/

/*
  This file is part of jtalk.

  jtalk is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  jtalk is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with jtalk; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <glib.h>

#include "jtalk.h"

typedef struct {
    GSource source;
    GPollFD poll;
} socket_fd_t;

static int sock;

void
socket_write(jtalk_client_t *client, const char *text)
{
    size_t l;
    ssize_t w;

    l = strlen(text) + 1;
    while(l > 0) {
	if((w = write(client->fd, text, l)) < 0) {
	    if(errno == EINTR)
		continue;
	    return; /* XXX */
	}
	l -= w;
	text += w;
    }
}

static gboolean
socket_fd_prepare(GSource *source, gint *timeout)
{
    socket_fd_t *sfd = (socket_fd_t *)source;
    (void)sfd;

    *timeout = -1;
    return(FALSE);
}

static gboolean
socket_fd_check(GSource *source)
{
    socket_fd_t *sfd = (socket_fd_t *)source;

    return((sfd->poll.revents & G_IO_IN) != 0);
}

static gboolean
socket_fd_dispatch(GSource *source, GSourceFunc cb, gpointer data)
{
    socket_fd_t *sfd = (socket_fd_t *)source;
    (void)sfd;

    return(cb(data));
}

static GSourceFuncs socket_fd_funcs = {
    .prepare = socket_fd_prepare,
    .check = socket_fd_check,
    .dispatch = socket_fd_dispatch,
    .finalize = NULL,
};

static socket_fd_t *
socket_fd_new(int fd)
{
    socket_fd_t *sfd;

    sfd = (socket_fd_t *)g_source_new(&socket_fd_funcs, sizeof(socket_fd_t));
    sfd->poll.fd = fd;
    sfd->poll.events = G_IO_IN;
    g_source_add_poll(&sfd->source, &sfd->poll);
    g_source_attach(&sfd->source, NULL);
    return(sfd);
}

static gboolean
socket_read(gpointer client_v)
{
    jtalk_client_t *client = client_v;
    ssize_t r;
    unsigned char *end;

    if(client->buffer_size - client->buffer_have < 1024) {
	client->buffer_size *= 2;
	client->buffer = g_realloc(client->buffer, client->buffer_size);
    }
    r = read(client->fd, client->buffer + client->buffer_have,
	client->buffer_size - client->buffer_have);
    if(r < 0) {
	if(errno == EINTR)
	    return(TRUE);
	return(FALSE);
    }
    if(r == 0)
	return(FALSE);
    client->buffer_have += r;
    while((end = memchr(client->buffer, 0, client->buffer_have)) != NULL) {
	jtalk_line_from_client(client, client->buffer);
	end++;
	client->buffer_have -= end - client->buffer;
	memmove(client->buffer, end, client->buffer_have);
    }
    return(TRUE);
}

static void
socket_kill(gpointer client_v)
{
    jtalk_client_t *client = client_v;

    close(client->fd);
    jtalk_client_closed(client);
}

static gboolean
socket_accept(gpointer dummy)
{
    (void)dummy;
    int fd;
    struct sockaddr_un raddr;
    socklen_t raddr_len = sizeof(raddr);
    socket_fd_t *sfd2;
    jtalk_client_t *client;

    if((fd = accept(sock, (struct sockaddr *)&raddr, &raddr_len)) < 0) {
	if(errno == EINTR)
	    return(TRUE);
	perror("accept");
	exit(1);
    }
    sfd2 = socket_fd_new(fd);
    client = g_new(jtalk_client_t, 1);
    jtalk_client_init(client);
    client->fd = fd;
    client->buffer_size = 4096;
    client->buffer = g_malloc(client->buffer_size);
    client->buffer_have = 0;
    client->source_id = g_source_get_id(&sfd2->source);
    g_source_set_callback(&sfd2->source,
	(GSourceFunc)socket_read, client, socket_kill);
    return(TRUE);
}

void
socket_server_start(void)
{
    struct sockaddr_un addr;
    socket_fd_t *s;
    char *psock;

    psock = jtalk_settings_get_file("socket/s");
    if((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
	perror("socket");
	exit(1);
    }
    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_UNIX;
    strcpy(addr.sun_path, psock);
    unlink(psock);
    free(psock);
    if(bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
	perror("bind");
	exit(1);
    }
    if(listen(sock, SOMAXCONN) < 0) {
	perror("listen");
	exit(1);
    }
    s = socket_fd_new(sock);
    g_source_set_callback(&s->source, (GSourceFunc)socket_accept, NULL, NULL);
}
